/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000-2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

/* ====================================================================
 * procrun.
 *
 * Contributed by Mladen Turk <mturk@mappingsoft.com>
 *
 * 05 Aug 2002
 * ==================================================================== 
 */

#ifndef STRICT
#define STRICT
#endif
#ifndef OEMRESOURCE
#define OEMRESOURCE
#endif

#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <objbase.h>
#include <shlobj.h>
#include <stdlib.h>
#include <stdio.h>

#include <Shlwapi.h>
#include <io.h>
#include <fcntl.h>
#include <process.h>
#include <time.h>
#include <stdarg.h>
#include <jni.h>
 
#include "procrun.h"

typedef HANDLE (__stdcall * PFNCREATERTHRD)(HANDLE, LPSECURITY_ATTRIBUTES, 
                                            DWORD, LPTHREAD_START_ROUTINE,
                                            LPVOID, DWORD, LPDWORD); 
typedef jint (JNICALL *JNI_GETDEFAULTJAVAVMINITARGS)(void *);
typedef jint (JNICALL *JNI_CREATEJAVAVM)(JavaVM **, JNIEnv **, void *);
typedef jint (JNICALL *JNI_GETCREATEDJAVAVMS)(JavaVM **, int, int *);

JNI_GETDEFAULTJAVAVMINITARGS jni_JNI_GetDefaultJavaVMInitArgs = NULL;
JNI_CREATEJAVAVM jni_JNI_CreateJavaVM = NULL;
JNI_GETCREATEDJAVAVMS jni_JNI_GetCreatedJavaVMs = NULL;

int report_service_status(DWORD, DWORD, DWORD, process_t *); 
static int g_proc_stderr_file = 0;
int g_proc_mode = 0;
/* The main envronment for services */
procrun_t *g_env = NULL;
static int g_is_windows_nt = 0;

#ifdef PROCRUN_WINAPP

extern DWORD WINAPI gui_thread(LPVOID param);

extern void ac_add_list_string(const char *str, int len);
extern int  ac_use_try;
extern int  ac_use_dlg;
extern int  ac_use_show;
extern RECT ac_winpos;
extern HINSTANCE ac_instance;
extern HWND ac_main_hwnd;
void ac_show_try_icon(HWND hwnd, DWORD message, const char *tip);

#endif

#ifdef _DEBUG
void dbprintf(char *format, ...)
{
    va_list args;
    char tid[4096 + 128];
    char buffer[4096];
    int len;

    if (!format) {
        len = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | 
                            FORMAT_MESSAGE_IGNORE_INSERTS,
                            NULL,
                            GetLastError(),
                            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                            buffer,
                            4096,
                            NULL);
    }
    else {
        va_start(args, format);
        len = _vsnprintf(buffer, 4096, format, args);
        va_end(args);
    }
    if (len > 0) {
        sprintf(tid, "[%04X] %s", GetCurrentThreadId(), buffer);
        if (g_proc_stderr_file > 0)
            write(g_proc_stderr_file, tid, strlen(tid));
        else
            fprintf(stderr,tid);
#ifdef _DEBUG_TRACE
        OutputDebugString(tid);
#endif
    }
}
#define DBPRINTF0(v1)               dbprintf(v1)
#define DBPRINTF1(v1, v2)           dbprintf(v1, v2)
#define DBPRINTF2(v1, v2, v3)       dbprintf(v1, v2, v3)
#else
#define DBPRINTF0(v1)
#define DBPRINTF1(v1, v2)
#define DBPRINTF2(v1, v2, v3)
#endif

static pool_t *pool_create()
{
    pool_t *pool = malloc(sizeof(pool_t));
    if (pool) {
        InitializeCriticalSection(&pool->lock);
        pool->size = 0;
        pool->mp[0].m = NULL;
        pool->mp[0].h = NULL;
    }
    return pool;
}

static int pool_destroy(pool_t *pool)
{
    int i = 0;
    for (i = 0; i < pool->size; i++) {
        if (pool->mp[i].m) {
            free(pool->mp[i].m);
        }
        if (pool->mp[i].h != INVALID_HANDLE_VALUE && pool->mp[i].h != NULL) {
            CloseHandle(pool->mp[i].h);
        }
    }
    DeleteCriticalSection(&pool->lock);
    free(pool);
    return i;
}

static void *pool_alloc(pool_t *pool, size_t size)
{
    void *m = malloc(size);
    EnterCriticalSection(&pool->lock);
    pool->mp[pool->size].m = INVALID_HANDLE_VALUE;
    pool->mp[pool->size++].m = m;
    LeaveCriticalSection(&pool->lock);
    return m;
}

static void *pool_calloc(pool_t *pool, size_t size)
{
    void *m = calloc(size, 1);
    EnterCriticalSection(&pool->lock);
    pool->mp[pool->size].h = INVALID_HANDLE_VALUE;
    pool->mp[pool->size++].m = m;
    LeaveCriticalSection(&pool->lock);
    return m;
}

static char *pool_strdup(pool_t *pool, const char *src)
{
    char *s = strdup(src);
    EnterCriticalSection(&pool->lock);
    pool->mp[pool->size].h = INVALID_HANDLE_VALUE;
    pool->mp[pool->size++].m = s;
    LeaveCriticalSection(&pool->lock);
    return s;
}

static void *pool_handle(pool_t *pool, HANDLE h)
{
    EnterCriticalSection(&pool->lock);
    pool->mp[pool->size].h = h;
    pool->mp[pool->size++].m = NULL;
    LeaveCriticalSection(&pool->lock);
    return h;
}

static void simple_encrypt(int seed, const char *str, unsigned char bytes[256])
{
    int i;
    char sc[256];

    srand(seed);
    memset(sc, 0, 256);
    strncpy(sc, str, 255);
    for (i = 0; i < 256; i ++) {
        bytes[i] = ((rand() % 256) ^ sc[i]);
    }
}

static void simple_decrypt(int seed, char *str, unsigned char bytes[256])
{
    int i;
    char sc[256];

    srand(seed);
    for (i = 0; i < 256; i ++) {
        sc[i] = ((rand() % 256) ^ bytes[i]);
    }
    strcpy(str, sc);
}

static void test_enc() 
{
    unsigned char b[256];
    char op[32];
    simple_encrypt(100, "test encryption", b);
    simple_decrypt(100, op, b);
    DBPRINTF1("ENC %s", op);
}

static void inject_exitprocess(PROCESS_INFORMATION *child) 
{
    PFNCREATERTHRD pfn_CreateRemoteThread;
    UINT exit_code = 2303;
    HANDLE rt = NULL, dup = NULL;
    DWORD  rtid, stat;
    BOOL isok;

    if (!child || !child->hProcess) {
        DBPRINTF0("No process\n");
        return;
    }
    if (!GetExitCodeProcess(child->hProcess, &stat) || 
        (stat != STILL_ACTIVE)) {
            child->hProcess = NULL;
            child->dwProcessId = 0;
            DBPRINTF1("The child process %d isn't active any more\n",
                     child->dwProcessId);
            return;
    } 
    GenerateConsoleCtrlEvent(CTRL_C_EVENT, child->dwProcessId);
    GenerateConsoleCtrlEvent(CTRL_BREAK_EVENT, child->dwProcessId);
#if 1
    Sleep(2000);
#endif
    if (!GetExitCodeProcess(child->hProcess, &stat) || 
        (stat != STILL_ACTIVE)) {
            child->hProcess = NULL;
            child->dwProcessId = 0;
            DBPRINTF0("Breaked by CTRL+C event\n");
            return;
    } 
    DBPRINTF1("Injecting ExitProcess to the child process %d\n",
             child->dwProcessId);

    pfn_CreateRemoteThread = (PFNCREATERTHRD)GetProcAddress(
                                             GetModuleHandle("KERNEL32.DLL"),
                                            "CreateRemoteThread");

    isok = DuplicateHandle(GetCurrentProcess(), 
                           child->hProcess, 
                           GetCurrentProcess(), 
                           &dup, 
                           PROCESS_ALL_ACCESS, 
                           FALSE, 0);
    isok = GetExitCodeProcess((isok) ? dup : child->hProcess, &stat);
    if (pfn_CreateRemoteThread) {        
        FARPROC pfnExitProc;
        if (isok && stat == STILL_ACTIVE) { 
           
            pfnExitProc = GetProcAddress(GetModuleHandle("KERNEL32.DLL"),
                                         "ExitProcess");
            rt = pfn_CreateRemoteThread(child->hProcess, 
                                        NULL, 
                                        0, 
                                        (LPTHREAD_START_ROUTINE)pfnExitProc,
                                        (PVOID)exit_code, 0, &rtid);
       }
    }
    else {
        if (isok && stat == STILL_ACTIVE) {
            DBPRINTF0("Could not CreateRemoteThread... Forcing TerminateProcess\n");
            TerminateProcess(child->hProcess, exit_code);
        }
    }

    if (rt) {
        if (WaitForSingleObject(child->hProcess, 2000) == WAIT_OBJECT_0) {
            CloseHandle(rt);
            DBPRINTF0("Exited cleanly\n");
        }
        else {
            DBPRINTF0("Forcing TerminateProcess\n");
            TerminateProcess(child->hProcess, exit_code);
        }
    }
    if (dup)
        CloseHandle(dup);
} 

int __cdecl compare(const void *arg1, const void *arg2)
{
    return _stricoll(*((char **)arg1),*((char **)arg2));
}

static char * merge_arrays(char **one, char **two, process_t *proc)
{
    int len = 0, n, cnt = 0;
    char *envp, *p;

    for (n = 0; one[n]; n++) {
        len += (strlen(one[n]) + 1);
        cnt++;
    }
    for (n = 0; two[n]; n++) {
        len += (strlen(two[n]) + 1);        
        cnt++;
    }
    p = envp = (char *)pool_calloc(proc->pool, len + 1);

    for (n = 0; one[n]; n++) {
        strcpy(p, one[n]);
        p += strlen(one[n]) + 1;
    }
    for (n = 0; two[n]; n++) {
        strcpy(p, two[n]);
        p += strlen(two[n]) + 1;
    }
    return envp;
}

static char * make_environment(char **envarr, char **envorg, process_t *proc)
{
    int len = 0, n, cnt = 0;
    char *envp, *p, **tmp;

    for (n = 0; envarr[n]; n++) {
        len += strlen(envarr[n]) + 1;
        cnt++;
    }
    for (n = 0; envorg[n]; n++) {
        len += strlen(envorg[n]) + 1;        
        cnt++;
    }
    p = envp = (char *)pool_calloc(proc->pool, len + 1);
    if (!p)
        return NULL;
    tmp = (char **)calloc(cnt + 1, sizeof(char *));
    cnt = 0;
    for (n = 0; envarr[n]; n++) {
            tmp[cnt++] = envarr[n];
    }
    for (n = 0; envorg[n]; n++) {
        if (STRN_COMPARE(envorg[n], PROCRUN_ENV_STDIN)) {}
        else if (STRN_COMPARE(envorg[n], PROCRUN_ENV_STDOUT)) {}
        else if (STRN_COMPARE(envorg[n], PROCRUN_ENV_STDERR)) {}
        else if (STRN_COMPARE(envorg[n], PROCRUN_ENV_PPID)) {}
        else
            tmp[cnt++] = envorg[n];
    }
    qsort((void *)tmp, (size_t)n, sizeof(char *), compare);
    for (n = 0; tmp[n]; n++) {
        strcpy(p, tmp[n]);
        p += strlen(tmp[n]) + 1;
    }
    free(tmp);
    return envp;
}

static int make_array(const char *str, char **arr, int size, process_t *proc)
{
    int i = 0;
    char *p;
    
    if (!str)
        return 0;
    for (p = (char *)str; p && *p; p++) {
        arr[i] = pool_strdup(proc->pool, p);
        while (*p)
            p++;
        i++;
        if (i + 1 > size)
            break;
    }
    return i;
}

/*  Parse command line argument.
 *  First command param starts with //name//value
 */
static int parse_args(int argc, char **argv, process_t *proc)
{
    int mode = 0;
    char *arg = argv[1];

    if ((strlen(arg) > 5) && arg[0] == '/' && arg[1] == '/') {
        if (STRN_COMPARE(arg, PROC_ARG_ENVPREFIX)) {
            proc->env_prefix = arg + STRN_SIZE(PROC_ARG_ENVPREFIX);
            mode = PROCRUN_CMD_ENVPREFIX;
        }
        else if (STRN_COMPARE(arg, PROC_ARG_RUN_JAVA)) {
            proc->java.path = pool_strdup(proc->pool, arg + 
                                          STRN_SIZE(PROC_ARG_RUN_JAVA));
            mode = PROCRUN_CMD_RUN_JAVA;
        }
        else if (STRN_COMPARE(arg, PROC_ARG_INSTALL_SERVICE)) {
            proc->service.name = pool_strdup(proc->pool, arg + 
                                          STRN_SIZE(PROC_ARG_INSTALL_SERVICE));
            mode = PROCRUN_CMD_INSTALL_SERVICE;
        }
        else if (STRN_COMPARE(arg, PROC_ARG_RUN_SERVICE)) {
            proc->service.name = pool_strdup(proc->pool, arg + 
                                             STRN_SIZE(PROC_ARG_RUN_SERVICE));
            mode = PROCRUN_CMD_RUN_SERVICE;
        }
        else if (STRN_COMPARE(arg, PROC_ARG_TEST_SERVICE)) {
            proc->service.name = pool_strdup(proc->pool, arg + 
                                             STRN_SIZE(PROC_ARG_TEST_SERVICE));
            mode = PROCRUN_CMD_TEST_SERVICE;
        }
#ifdef PROCRUN_WINAPP
        else if (STRN_COMPARE(arg, PROC_ARG_GUIT_SERVICE)) {
            proc->service.name = pool_strdup(proc->pool, arg + 
                                             STRN_SIZE(PROC_ARG_GUIT_SERVICE));
            mode = PROCRUN_CMD_GUIT_SERVICE;
            ac_use_try = 1;
            ac_use_dlg = 1;
        }
        else if (STRN_COMPARE(arg, PROC_ARG_GUID_SERVICE)) {
            proc->service.name = pool_strdup(proc->pool, arg + 
                                             STRN_SIZE(PROC_ARG_GUID_SERVICE));
            mode = PROCRUN_CMD_GUID_SERVICE;
            ac_use_dlg = 1;
            ac_use_show = 1;
        }
        else if (STRN_COMPARE(arg, PROC_ARG_GUID_PROCESS)) {
            mode = PROCRUN_CMD_GUID_PROCESS;
            ac_use_dlg = 1;
            ac_use_show = 1;
        }
#endif
        else if (STRN_COMPARE(arg, PROC_ARG_STOP_SERVICE)) {
            proc->service.name = pool_strdup(proc->pool, arg +
                                             STRN_SIZE(PROC_ARG_STOP_SERVICE));
            mode = PROCRUN_CMD_STOP_SERVICE;
        }
        else if (STRN_COMPARE(arg, PROC_ARG_DELETE_SERVICE)) {
            proc->service.name = pool_strdup(proc->pool, arg +
                                           STRN_SIZE(PROC_ARG_DELETE_SERVICE));
            mode = PROCRUN_CMD_DELETE_SERVICE;
        }
        else if (STRN_COMPARE(arg, PROC_ARG_UPDATE_SERVICE)) {
            proc->service.name = pool_strdup(proc->pool, arg +
                                           STRN_SIZE(PROC_ARG_UPDATE_SERVICE));
            mode = PROCRUN_CMD_UPDATE_SERVICE;
        }
    }
    return mode;
}

static void debug_process(int argc, char **argv, process_t *p)
{
    DBPRINTF1("DUMPING %s\n", argv[0]);
    DBPRINTF0("  SERVICE:\n");
    DBPRINTF1("    name        : %s\n", p->service.name);
    DBPRINTF1("    description : %s\n", p->service.description);
    DBPRINTF1("    path        : %s\n", p->service.path);
    DBPRINTF1("    image       : %s\n", p->service.image);
    DBPRINTF1("    infile      : %s\n", p->service.inname);
    DBPRINTF1("    outfile     : %s\n", p->service.outname);
    DBPRINTF1("    errfile     : %s\n", p->service.errname);
    DBPRINTF1("    argvw       : %s\n", p->argw);
    DBPRINTF0("  JAVAVM:\n");
    DBPRINTF1("    name        : %s\n", p->java.path);
    DBPRINTF1("    start       : %s\n", p->java.start_class);
    DBPRINTF1("    stop        : %s\n", p->java.stop_class);
    DBPRINTF1("    start method: %s\n", p->java.start_method);
    DBPRINTF1("    stop method : %s\n", p->java.stop_method);
    DBPRINTF1("    start param : %s\n", p->java.start_param);
    DBPRINTF1("    stop param  : %s\n", p->java.stop_param);

    DBPRINTF0("DONE...\n");
}

static int procrun_addenv(char *name, char *value, int val, process_t *proc)
{
    int i;

    for (i = 0; i < PROC_ENV_COUNT; i++)
        if (proc->env[i] == NULL)
            break;
    if (i < PROC_ENV_COUNT) {
        if (value) {
            proc->env[i] = (char *)pool_alloc(proc->pool, strlen(name) + 
                                              strlen(value) + 1);
            strcpy(proc->env[i], name);
            strupr(proc->env[i]);
            strcat(proc->env[i], value);
        }
        else {
            proc->env[i] = (char *)pool_alloc(proc->pool, strlen(name) + 34);
            strcpy(proc->env[i], name);
            strupr(proc->env[i]);
            itoa(val, proc->env[i] + strlen(name), 10);
        }
        return i;
    }
    else
        return -1;
}

static int procrun_readenv(process_t *proc, char **envp)
{
    int i, rv = 0;
    char *env;
    HANDLE h;

    for (i = 0; envp[i]; i++) {
        env = envp[i];
        if (STRN_COMPARE(env, PROCRUN_ENV_STDIN)) {
            h = (HANDLE)atoi(env + STRN_SIZE(PROCRUN_ENV_STDIN));
            if (!DuplicateHandle(proc->pinfo.hProcess, 
                                 h,
                                 proc->pinfo.hProcess,
                                 &proc->h_stdin[0], 
                                 0L, TRUE, DUPLICATE_SAME_ACCESS)) {
                DBPRINTF1("%s\tDuplicateHandle for STDIN failed\n", env[i]);
                proc->h_stdin[0] = INVALID_HANDLE_VALUE;
                return 0;
            }
            CloseHandle(h);
            ++rv;
        }
        else if (STRN_COMPARE(env, PROCRUN_ENV_STDOUT)) {
            h = (HANDLE)atoi(env + STRN_SIZE(PROCRUN_ENV_STDOUT));
            if (!DuplicateHandle(proc->pinfo.hProcess, 
                                 h,
                                 proc->pinfo.hProcess,
                                 &proc->h_stdout[1], 
                                 0L, TRUE, DUPLICATE_SAME_ACCESS)) {
                DBPRINTF1("%s\tDuplicateHandle for STDOUT failed\n", env[i]);
                proc->h_stdout[1] = INVALID_HANDLE_VALUE;
                return 0;
            }
            CloseHandle(h);
            ++rv;
        }
        else if (STRN_COMPARE(env, PROCRUN_ENV_STDERR)) {
            h = (HANDLE)atoi(env + STRN_SIZE(PROCRUN_ENV_STDERR));
            if (!DuplicateHandle(proc->pinfo.hProcess, 
                                 h,
                                 proc->pinfo.hProcess,
                                 &proc->h_stderr[1], 
                                 0L, TRUE, DUPLICATE_SAME_ACCESS)) {
                DBPRINTF1("%s\tDuplicateHandle for STDERR failed\n", env[i]);
                proc->h_stderr[1] = INVALID_HANDLE_VALUE;
                return 0;
            }
            CloseHandle(h);
            ++rv;
        }
        else if (STRN_COMPARE(env, PROCRUN_ENV_ERRFILE)) {
            h = (HANDLE)atoi(env + STRN_SIZE(PROCRUN_ENV_ERRFILE));
            g_proc_stderr_file = _open_osfhandle((long)h,
                                                  _O_APPEND | _O_TEXT);
        }
        else if (STRN_COMPARE(env, PROCRUN_ENV_PPID)) {
            proc->ppid = atoi(env + STRN_SIZE(PROCRUN_ENV_PPID));
        }

    }
    return rv;
}

static char* procrun_guess_jvm(process_t *proc)
{
    HKEY hkjs;
    char jvm[MAX_PATH+1];
    char reg[MAX_PATH+1];
    char *cver;
    unsigned long err, klen = MAX_PATH;
    
    strcpy(reg, JAVASOFT_REGKEY);
    cver = &reg[sizeof(JAVASOFT_REGKEY)-1];

    if ((err = RegOpenKeyEx(HKEY_LOCAL_MACHINE, reg,
                            0, KEY_READ, &hkjs)) != ERROR_SUCCESS) {
       DBPRINTF0("procrun_guess_jvm() failed to open Registry key\n");
       return NULL;
    }
    if ((err = RegQueryValueEx(hkjs, "CurrentVersion", NULL, NULL, 
                               (unsigned char *)cver,
                               &klen)) != ERROR_SUCCESS) {
        DBPRINTF0("procrun_guess_jvm() failed obtaining Current Java Version\n");
        RegCloseKey(hkjs);
        return NULL;
    }
    RegCloseKey(hkjs);
    
    if ((err = RegOpenKeyEx(HKEY_LOCAL_MACHINE, reg,
                            0, KEY_READ, &hkjs) ) != ERROR_SUCCESS) {
        DBPRINTF1("procrun_guess_jvm() failed to open Registry key %s\n", reg);
        return NULL;
    }
    klen = MAX_PATH;
    if ((err = RegQueryValueEx(hkjs, "RuntimeLib", NULL, NULL, 
                               (unsigned char *)jvm, 
                               &klen)) != ERROR_SUCCESS) {
        DBPRINTF0("procrun_guess_jvm() failed obtaining Runtime Library\n");
        RegCloseKey(hkjs);
        return NULL;
    }
    RegCloseKey(hkjs);

    return pool_strdup(proc->pool, jvm);
} 

static char* procrun_guess_java(process_t *proc, const char *image)
{
    HKEY hkjs;
    char jbin[MAX_PATH+1];
    char reg[MAX_PATH+1];
    char *cver;
    unsigned long err, klen = MAX_PATH;
    
    strcpy(reg, JAVASOFT_REGKEY);
    cver = &reg[sizeof(JAVASOFT_REGKEY)-1];

    if ((err = RegOpenKeyEx(HKEY_LOCAL_MACHINE, reg,
                            0, KEY_READ, &hkjs)) != ERROR_SUCCESS) {
       DBPRINTF0("procrun_guess_jvm() failed to open Registry key\n");
       return NULL;
    }
    if ((err = RegQueryValueEx(hkjs, "CurrentVersion", NULL, NULL, 
                               (unsigned char *)cver,
                               &klen)) != ERROR_SUCCESS) {
        DBPRINTF0("procrun_guess_jvm() failed obtaining Current Java Version\n");
        RegCloseKey(hkjs);
        return NULL;
    }
    RegCloseKey(hkjs);
    
    if ((err = RegOpenKeyEx(HKEY_LOCAL_MACHINE, reg,
                            0, KEY_READ, &hkjs) ) != ERROR_SUCCESS) {
        DBPRINTF1("procrun_guess_jvm() failed to open Registry key %s\n", reg);
        return NULL;
    }
    klen = MAX_PATH;
    if ((err = RegQueryValueEx(hkjs, "JavaHome", NULL, NULL, 
                               (unsigned char *)jbin, 
                               &klen)) != ERROR_SUCCESS) {
        DBPRINTF0("procrun_guess_jvm() failed obtaining Runtime Library\n");
        RegCloseKey(hkjs);
        return NULL;
    }
    RegCloseKey(hkjs);
    strcat(jbin, "\\bin\\");
    strcat(jbin, image);
    strcat(jbin, ".exe");
    return pool_strdup(proc->pool, jbin);
} 

static int procrun_service_params(process_t *proc)
{
    HKEY key;
    char skey[256];
    char kval[MAX_PATH];
    unsigned long klen;
    DWORD err;
    
    if (!proc->service.name)
        return 0;

    sprintf(skey, PROCRUN_REGKEY_RPARAMS, proc->service.name);
    DBPRINTF1("getting %s\n", skey);
    if ((err = RegOpenKeyEx(HKEY_LOCAL_MACHINE, skey,
                            0, KEY_READ, &key)) == ERROR_SUCCESS) {

        klen = MAX_PATH;
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_DESCRIPTION, NULL, NULL, 
                                   (unsigned char *)kval,
                                   &klen)) == ERROR_SUCCESS) {
            proc->service.description = pool_strdup(proc->pool, kval);
        }
        klen = MAX_PATH;
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_DISPLAY, NULL, NULL, 
                                   (unsigned char *)kval,
                                   &klen)) == ERROR_SUCCESS) {
            proc->service.display = pool_strdup(proc->pool, kval);
        }
        klen = MAX_PATH;
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_WORKPATH, NULL, NULL, 
                                   (unsigned char *)kval,
                                   &klen)) == ERROR_SUCCESS) {
            proc->service.path = pool_strdup(proc->pool, kval);
        }
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_IMAGE, NULL, NULL, 
                                   NULL,
                                   &klen)) == ERROR_SUCCESS) {
            proc->service.image = (char *)pool_alloc(proc->pool, klen);
            if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_IMAGE, NULL, NULL, 
                                   (unsigned char *)proc->service.image,
                                   &klen)) != ERROR_SUCCESS) {
                proc->service.image = NULL;
            }
        }
        klen = MAX_PATH;
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_ACCOUNT, NULL, NULL, 
                                   (unsigned char *)kval,
                                   &klen)) == ERROR_SUCCESS) {
            proc->service.account = pool_strdup(proc->pool, kval);
        }
        klen = MAX_PATH;
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_PASSWORD, NULL, NULL, 
                                   (unsigned char *)kval,
                                   &klen)) == ERROR_SUCCESS) {
            proc->service.password = pool_calloc(proc->pool, 256);
            simple_decrypt(100, proc->service.password,  kval);
        }
        klen = MAX_PATH;
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_STARTCLASS, NULL, NULL, 
                                   (unsigned char *)kval,
                                   &klen)) == ERROR_SUCCESS) {
            char *p;
            p = strchr(kval, ';');
            if (p) *p = '\0';
            proc->java.start_class = pool_strdup(proc->pool, kval);
            if (p) {
                ++p;
                proc->java.start_method = pool_strdup(proc->pool, p);
                p = strchr(proc->java.start_method, ';');
                if (p) {
                    *p = '\0';
                    ++p;
                    proc->java.start_param = pool_strdup(proc->pool, p);
                }
            }
        }
        klen = MAX_PATH;
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_STOPCLASS, NULL, NULL, 
                                   (unsigned char *)kval,
                                   &klen)) == ERROR_SUCCESS) {
            char *p;
            p = strchr(kval, ';');
            if (p) *p = '\0';
            proc->java.stop_class = pool_strdup(proc->pool, kval);
            if (p) {
                ++p;
                proc->java.stop_method = pool_strdup(proc->pool, p);
                p = strchr(proc->java.stop_method, ';');
                if (p) {
                    *p = '\0';
                    ++p;
                    proc->java.stop_param = pool_strdup(proc->pool, p);
                }
            }
        }
        klen = MAX_PATH;
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_STDINFILE, NULL, NULL, 
                                   (unsigned char *)kval,
                                   &klen)) == ERROR_SUCCESS) {
            proc->service.infile = CreateFile(kval,
                                              GENERIC_READ,
                                              0,
                                              NULL,
                                              OPEN_EXISTING,
                                              FILE_ATTRIBUTE_NORMAL,
                                              NULL);
            if (proc->service.infile != INVALID_HANDLE_VALUE) {
                if (proc->h_stdin[1] != INVALID_HANDLE_VALUE)
                    CloseHandle(proc->h_stdin[1]);
                proc->h_stdin[1] = proc->service.infile;
                proc->service.inname = pool_strdup(proc->pool, kval);
            }
            pool_handle(proc->pool, proc->service.infile);
        }
        klen = MAX_PATH;
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_STDOUTFILE, NULL, NULL, 
                                   (unsigned char *)kval,
                                   &klen)) == ERROR_SUCCESS) {
            proc->service.outfile = CreateFile(kval,
                                               GENERIC_WRITE,
                                               FILE_SHARE_READ,
                                               NULL,
                                               OPEN_ALWAYS,
                                               FILE_ATTRIBUTE_NORMAL,
                                               NULL);
            if (proc->service.outfile != INVALID_HANDLE_VALUE) {
                SetFilePointer(proc->service.outfile, 0L, NULL, FILE_END);
                if (proc->h_stdout[1] != INVALID_HANDLE_VALUE)
                    CloseHandle(proc->h_stdout[1]);
                proc->h_stdout[1] = proc->service.outfile;
                proc->service.outname = pool_strdup(proc->pool, kval);
            }
            pool_handle(proc->pool, proc->service.outfile);
        }
        klen = MAX_PATH;
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_STDERRFILE, NULL, NULL, 
                                   (unsigned char *)kval,
                                   &klen)) == ERROR_SUCCESS) {
            proc->service.errfile = CreateFile(kval,
                                               GENERIC_WRITE,
                                               FILE_SHARE_READ,
                                               NULL,
                                               OPEN_ALWAYS,
                                               FILE_ATTRIBUTE_NORMAL,
                                               NULL);
            if (proc->service.errfile != INVALID_HANDLE_VALUE) {
                SetFilePointer(proc->service.errfile, 0L, NULL, FILE_END);
                if (proc->h_stderr[1] != INVALID_HANDLE_VALUE)
                    CloseHandle(proc->h_stderr[1]);
                proc->h_stderr[1] = proc->service.errfile;
                proc->service.errname = pool_strdup(proc->pool, kval);
            }
            pool_handle(proc->pool, proc->service.errfile);
        }
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_CMDARGS, NULL, NULL, 
                                   NULL,
                                   &klen)) == ERROR_SUCCESS) {
            proc->argw = (char *)pool_alloc(proc->pool, klen);
            if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_CMDARGS, NULL, NULL, 
                                       (unsigned char *)proc->argw,
                                       &klen)) != ERROR_SUCCESS) {
                proc->argw = NULL;
            }
            
        }
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_JVM_OPTS, NULL, NULL, 
                                   NULL,
                                   &klen)) == ERROR_SUCCESS) {
            proc->java.opts = (char *)pool_alloc(proc->pool, klen);
            if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_JVM_OPTS, NULL, NULL, 
                                       (unsigned char *)proc->java.opts,
                                       &klen)) != ERROR_SUCCESS) {
                proc->java.opts = NULL;
            }
        }
#ifdef PROCRUN_WINAPP
        klen = MAX_PATH;
        if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_WINPOS, NULL, NULL, 
                                   (unsigned char *)kval,
                                   &klen)) == ERROR_SUCCESS) {
            sscanf(kval, "%d %d %d %d", &ac_winpos.left, &ac_winpos.right,
                                        &ac_winpos.top, &ac_winpos.bottom);
        }
#endif
        RegCloseKey(key); 
        return 0;
    }
    else
        return -1;
}

static int procrun_load_jvm(process_t *proc, int mode)
{
    int has_java = 0;
    char jvm_path[MAX_PATH + 1];
    
    if (!proc->java.start_class) {
        return 0;
    }
    jvm_path[0] = '\0';
    switch (mode) {
        case PROCRUN_CMD_RUN_JAVA:
            has_java = 1;
            break;
        case PROCRUN_CMD_RUN_SERVICE:
        case PROCRUN_CMD_TEST_SERVICE:
        case PROCRUN_CMD_GUIT_SERVICE:
        case PROCRUN_CMD_GUID_SERVICE:
            {
                HKEY key;
                char skey[256];
                unsigned long klen = MAX_PATH;
                DWORD err;
                sprintf(skey, PROCRUN_REGKEY_RPARAMS, proc->service.name);
                if ((err = RegOpenKeyEx(HKEY_LOCAL_MACHINE, skey,
                                        0, KEY_READ, &key)) == ERROR_SUCCESS) {
                    
                    if ((err = RegQueryValueEx(key, PROCRUN_PARAMS_JVM,
                                               NULL, NULL, 
                                               (unsigned char *)jvm_path,
                                               &klen)) == ERROR_SUCCESS) {
                        has_java = 1;
                    }
                    RegCloseKey(key); 
                }
            }
            break;
    }
    if (has_java) {
        UINT em;
        if (strlen(jvm_path)) {
            if (strlen(jvm_path) < 6)
                proc->java.path = procrun_guess_jvm(proc);
            else
                proc->java.path = pool_strdup(proc->pool, jvm_path);
            if (strnicmp(jvm_path, "java", 4) == 0) {
                proc->java.jpath = procrun_guess_java(proc, jvm_path);
            }
        }
        DBPRINTF1("jvm dll path %s", proc->java.path);
        DBPRINTF1("java    path %s", proc->java.jpath);
        if (!proc->java.path || !proc->java.start_method)
            return -1;
        /* Try to load the jvm dll */ 
        em = SetErrorMode(SEM_FAILCRITICALERRORS);
        proc->java.dll = LoadLibraryEx(proc->java.path, NULL, 0); 
        if (!proc->java.dll)
            proc->java.dll = LoadLibraryEx(proc->java.path, NULL,
                                           LOAD_WITH_ALTERED_SEARCH_PATH);
        SetErrorMode(em);
        if (!proc->java.dll)
            return -1;
        /* resolve symbols */
        jni_JNI_GetDefaultJavaVMInitArgs = (JNI_GETDEFAULTJAVAVMINITARGS)
                                            GetProcAddress(proc->java.dll,
                                            "JNI_GetDefaultJavaVMInitArgs");
        jni_JNI_CreateJavaVM = (JNI_CREATEJAVAVM)
                               GetProcAddress(proc->java.dll,
                               "JNI_CreateJavaVM");
        jni_JNI_GetCreatedJavaVMs = (JNI_GETCREATEDJAVAVMS)
                                    GetProcAddress(proc->java.dll,
                                    "JNI_GetCreatedJavaVMs");
        if (jni_JNI_GetDefaultJavaVMInitArgs == NULL ||
            jni_JNI_CreateJavaVM == NULL ||
            jni_JNI_GetCreatedJavaVMs == NULL) {
            FreeLibrary(proc->java.dll);
            proc->java.dll = NULL;
            return -1;
        }
        DBPRINTF1("JVM %s Loaded\n", proc->java.path);
    }
    return 0;
}

/* JVM hooks */
static int jni_exit_signaled = 0;
static int jni_exit_code = 0;
static int jni_abort_signaled = 0;

static void jni_exit_hook(int code)
{
    jni_exit_signaled = -1;
#if 1
    jni_abort_signaled = -1;    
#endif
    jni_exit_code = code;

    DBPRINTF1("JVM exit hook called %d\n", code);
}

static void jni_abort_hook()
{
    jni_abort_signaled = -1;    

    DBPRINTF0("JVM abort hook called\n");
}  

static JNIEnv *jni_attach(process_t *proc)
{
    JNIEnv *env = NULL;
    int err;
    JavaVM *jvm = proc->java.jvm;

    if (jvm == NULL || jni_abort_signaled)
        return NULL;
    err = (*jvm)->GetEnv(jvm, (void **)&env, JNI_VERSION_1_2);  
    if (err == 0)
        return env;        
    if (err != JNI_EDETACHED) {
        return NULL;
    }
    err = (*jvm)->AttachCurrentThread(jvm,
                                      (void **)&env,
                                      NULL);
    if (err != 0) {
        return NULL;
    }
    return env;
}

static int jni_detach(process_t *proc)
{
    JavaVM *jvm = proc->java.jvm;

    if (jvm == NULL || jni_abort_signaled)
        return -1;
    return (*jvm)->DetachCurrentThread(jvm);
}

static int procrun_destroy_jvm(process_t *proc)
{
    JavaVM *jvm = proc->java.jvm;
    int err;
    JNIEnv *env;

    if (!proc->java.dll || !jvm)
        return 0;
    env = jni_attach(proc);
    if (!env)
        goto cleanup;
    if (proc->java.stop_bridge && proc->java.stop_mid) {
        jclass strclass;
        jarray jargs = NULL;
        DBPRINTF1("Calling shutdown %s", proc->java.stop_class); 

        strclass = (*env)->FindClass(env, "java/lang/String");
        if (proc->java.stop_param) {
            jstring arg = (*env)->NewStringUTF(env, proc->java.stop_param);
            jargs = (*env)->NewObjectArray(env, 1, strclass, NULL);
            (*env)->SetObjectArrayElement(env, jargs, 0, arg);
        }
        (*env)->CallStaticVoidMethod(env,
                                     proc->java.stop_bridge,
                                     proc->java.stop_mid,
                                     jargs);
    }
    else if (!proc->java.jpath) {
        /* Call java.lang.System.exit(0) */
        jclass sysclass;
        jmethodID exitid;
        sysclass = (*env)->FindClass(env, "java/lang/System");
        if (!sysclass)
            goto cleanup;
        exitid   = (*env)->GetStaticMethodID(env, sysclass, "exit", "(I)V");
        if (!exitid)
            goto cleanup;
        report_service_status(SERVICE_STOPPED, 0, 0,
                              g_env->m);
        DBPRINTF0("Forcing shutdown using System.exit(0)");
 
        (*env)->CallStaticVoidMethod(env, sysclass, exitid, 0);
    }
cleanup:
    err = (*jvm)->DestroyJavaVM(jvm);
    FreeLibrary(proc->java.dll);
    proc->java.dll = NULL;
    return err;
}

static int procrun_init_jvm(process_t *proc)
{
    int jvm_version;
    JDK1_1InitArgs vm_args11;
    JavaVMInitArgs vm_args;
    JavaVMOption options[32];
    JNIEnv *env;
    JavaVM *jvm;
    jclass strclass;
    jarray jargs = NULL;
    char *cp;
    char *opts[32];
    int optn, i, err;

    vm_args11.version = JNI_VERSION_1_2;
     
    DBPRINTF0("Initializing JVM\n");
    if (jni_JNI_GetDefaultJavaVMInitArgs(&vm_args11) != 0) {
        DBPRINTF0("Could not find Default InitArgs\n");
        return -1;
    }
    jvm_version= vm_args11.version;

    if (jvm_version != JNI_VERSION_1_2) {
        DBPRINTF1("Found: %X expecting 1.2 Java Version\n", jvm_version);
        return -1;
    }
    
    optn = make_array(proc->java.opts, opts, 30, proc);
    for (i = 0; i < optn; i++)
        options[i].optionString = opts[i];
    cp = (char *)pool_alloc(proc->pool, strlen("-Djava.class.path=") +
                            strlen(proc->service.image) + 1);
    strcpy(cp, "-Djava.class.path=");
    strcat(cp, proc->service.image);
    options[optn++].optionString = cp;
    pool_alloc(proc->pool, 100);
    DBPRINTF1("-Djava.class.path=%s", proc->service.image);
    /* Set the abort and exit hooks */
#if 0
    options[optn].optionString = "exit";
    options[optn++].extraInfo = jni_exit_hook;
#endif
    options[optn].optionString = "abort";
    options[optn++].extraInfo = jni_abort_hook; 

    for (i = 0; i < optn; i++)
    DBPRINTF2("OPT %d %s", i , options[i].optionString);

    vm_args.version = JNI_VERSION_1_2;
    vm_args.options = options;
    vm_args.nOptions = optn;    
    vm_args.ignoreUnrecognized = JNI_TRUE;
 
    err = jni_JNI_CreateJavaVM(&jvm, &env, &vm_args);
    if (err == JNI_EEXIST) {       
        int vmcount;
        
        jni_JNI_GetCreatedJavaVMs(&jvm, 1, &vmcount);        
        if (jvm == NULL) {
            DBPRINTF0("Error creating JVM\n");
            return -1;
        }
    }
    proc->java.jvm = jvm;
    for (i = 0; i < (int)strlen(proc->java.start_class); i++) {
        if (proc->java.start_class[i] == '.')
            proc->java.start_class[i] = '/';
    }
    proc->java.start_bridge = (*env)->FindClass(env, proc->java.start_class);
    if (!proc->java.start_bridge) {
        DBPRINTF1("Couldn't find Startup class %s\n", proc->java.start_class);
        goto cleanup;
    }
    proc->java.start_mid = (*env)->GetStaticMethodID(env, proc->java.start_bridge,
                                                     proc->java.start_method, 
                                                     "([Ljava/lang/String;)V");

    if (!proc->java.start_mid) {
        DBPRINTF1("Couldn't find Startup class method %s\n", proc->java.start_method);
        goto cleanup;
    }
    if (proc->java.stop_class && proc->java.stop_method) {
        for (i = 0; i < (int)strlen(proc->java.stop_class); i++) {
            if (proc->java.stop_class[i] == '.')
                proc->java.stop_class[i] = '/';
        }
        proc->java.stop_bridge = (*env)->FindClass(env, proc->java.stop_class);
        if (!proc->java.start_bridge) {
            goto cleanup;
        }    
        proc->java.stop_mid = (*env)->GetStaticMethodID(env, proc->java.stop_bridge,
                                                        proc->java.stop_method, 
                                                        "([Ljava/lang/String;)V");
    }
    
    /* check if we have java.exe as worker process 
     * in that case don't call the startup class. 
     */
    if (proc->java.jpath != NULL) {
        return 0;
    } 
    strclass = (*env)->FindClass(env, "java/lang/String");
    if (proc->java.start_param) {
        jstring arg = (*env)->NewStringUTF(env, proc->java.start_param);
        jargs = (*env)->NewObjectArray(env, 1, strclass, NULL);
        (*env)->SetObjectArrayElement(env, jargs, 0, arg);
    }
    (*env)->CallStaticVoidMethod(env,
                                 proc->java.start_bridge,
                                 proc->java.start_mid,
                                 jargs);

    DBPRINTF1("JVM Main class %s finished\n", proc->java.start_class);
    return 0;
cleanup:
    if (proc->java.jvm) {
        (*(proc->java.jvm))->DestroyJavaVM(proc->java.jvm);
        FreeLibrary(proc->java.dll);
        proc->java.dll = NULL;
        proc->java.jvm = NULL;
    }
    return -1;
}


DWORD WINAPI wait_thread(LPVOID param)
{
    procrun_t *env = (procrun_t *)param;
    
    Sleep(1000);
    WaitForSingleObject(env->c->pinfo.hThread, INFINITE);
    CloseHandle(env->c->pinfo.hThread);
    env->c->pinfo.hThread = NULL;
    env->c->pinfo.dwProcessId = 0;
    SetEvent(env->m->events[1]);

    return 0;
}  

DWORD WINAPI stdout_thread(LPVOID param)
{
    unsigned char  ch;
    DWORD readed, written;
    procrun_t *env = (procrun_t *)param;
#ifdef PROCRUN_WINAPP
    static unsigned char buff[MAX_PATH + 1];
    int n = 0;
#endif

    while (env->c->h_stdout[3] && 
           (ReadFile(env->c->h_stdout[3], &ch, 1, &readed, NULL) == TRUE))  {
        if (readed) {
#ifdef PROCRUN_WINAPP
            if (ac_use_dlg) {
                if (ch == '\n' || n >= MAX_PATH) {
                    buff[n] = '\0';
                    DBPRINTF1("RD %s", buff);
                    ac_add_list_string(buff, n);
                    n  = 0;
                } 
                else if (ch == '\t' && n < (MAX_PATH - 4)) {
                    int i;
                    for (i = 0; i < 4; ++i)
                        buff[n++] = ' ';
                }
                else if (ch != '\r')
                    buff[n++] = ch;
                else if (ch != '\b')
                    buff[n++] = ' ';
                SwitchToThread();
            }
#endif
            if (WriteFile(env->m->h_stdout[0], &ch, 1, &written, NULL) == TRUE) {
                SwitchToThread();
            }
            else
                break;
            readed = 0;
        }
    }
    SetEvent(env->m->events[2]);
    return 0;
}  

DWORD WINAPI stderr_thread(LPVOID param)
{
    unsigned char  ch;
    DWORD readed, written;
    procrun_t *env = (procrun_t *)param;
#ifdef PROCRUN_WINAPP
    static unsigned char buff[MAX_PATH + 1];
    int n = 0;
#endif

    while (env->c->h_stderr[3] && 
          (ReadFile(env->c->h_stderr[3], &ch, 1, &readed, NULL) == TRUE))  {
        if (readed) {
#ifdef PROCRUN_WINAPP
            if (ac_use_dlg) {
                if (ch == '\n' || n >= MAX_PATH) {
                    buff[n] = '\0';
                    DBPRINTF1("RD %s", buff);
                    ac_add_list_string(buff, n);
                    n  = 0;
                } 
                else if (ch == '\t' && n < (MAX_PATH - 4)) {
                    int i;
                    for (i = 0; i < 4; ++i)
                        buff[n++] = ' ';
                }
                else if (ch != '\r')
                    buff[n++] = ch;
                else if (ch != '\b')
                    buff[n++] = ' ';
                SwitchToThread();
            }
#endif
            if (WriteFile(env->m->h_stderr[0], &ch, 1, &written, NULL) == TRUE) {
                SwitchToThread();
            }
            else
                break;
            readed = 0;
        }
    }
    SetEvent(env->m->events[3]);
    return 0;
}  

static int procrun_create_pipes(procrun_t *env)
{
    SECURITY_ATTRIBUTES sa;    
   
    sa.nLength = sizeof(SECURITY_ATTRIBUTES);
    sa.bInheritHandle = TRUE;
    sa.lpSecurityDescriptor = NULL;

    /* redirect stdout */
    if (env->m->h_stdout[1] == INVALID_HANDLE_VALUE)
        env->m->h_stdout[0] = GetStdHandle(STD_OUTPUT_HANDLE);
    else
        env->m->h_stdout[0] = env->m->h_stdout[1];

    if (!CreatePipe(&env->c->h_stdout[0],
                    &env->c->h_stdout[1], &sa, 0)) {
            return -1;
    }
    SetStdHandle(STD_OUTPUT_HANDLE, env->c->h_stdout[1]);
    pool_handle(env->c->pool, env->c->h_stdout[1]);

    if (!DuplicateHandle(env->m->pinfo.hProcess,
                         env->c->h_stdout[0],
                         env->m->pinfo.hProcess,
                         &env->c->h_stdout[3], 
                         0, FALSE, DUPLICATE_SAME_ACCESS)) {
        return -1;
    }
    CloseHandle(env->c->h_stdout[0]);
    pool_handle(env->c->pool, env->c->h_stdout[3]);
    
    /* redirect stderr */
    if (env->m->h_stderr[1] == INVALID_HANDLE_VALUE)
        env->m->h_stderr[0] = GetStdHandle(STD_ERROR_HANDLE);
    else
        env->m->h_stderr[0] = env->m->h_stderr[1];

    if (!CreatePipe(&env->c->h_stderr[0],
                    &env->c->h_stderr[1], &sa, 0)) {
            return -1;
    }
    SetStdHandle(STD_ERROR_HANDLE, env->c->h_stderr[1]);
    pool_handle(env->c->pool, env->c->h_stderr[1]);

    if (!DuplicateHandle(env->m->pinfo.hProcess,
                         env->c->h_stderr[0],
                         env->m->pinfo.hProcess,
                         &env->c->h_stderr[3], 
                         0, FALSE, DUPLICATE_SAME_ACCESS)) {
        return -1;
    }
    CloseHandle(env->c->h_stderr[0]);
    pool_handle(env->c->pool, env->c->h_stderr[3]);

    /* redirect stdin */
    if (env->m->h_stdin[1] == INVALID_HANDLE_VALUE)
        env->m->h_stdin[0] = GetStdHandle(STD_INPUT_HANDLE);
    else
        env->m->h_stdin[0] = env->m->h_stdin[1];

    if (!CreatePipe(&env->c->h_stdin[0],
                    &env->c->h_stdin[1], &sa, 0)) {
            return -1;
    }
    SetStdHandle(STD_INPUT_HANDLE, env->c->h_stdin[0]);  
    pool_handle(env->c->pool, env->c->h_stdin[0]);
    if (!DuplicateHandle(env->m->pinfo.hProcess,
                         env->c->h_stdin[1],
                         env->m->pinfo.hProcess,
                         &env->c->h_stdin[3], 
                         0, FALSE, DUPLICATE_SAME_ACCESS)) {
        return -1;
    }
    
    CloseHandle(env->c->h_stdin[1]);
    pool_handle(env->c->pool, env->c->h_stdin[3]);

    return 0;
}

static int procrun_write_stdin(procrun_t *env)
{
    DWORD readed, written; 
    char buf[PROC_BUFSIZE]; 
    
    if (!env->m->service.infile)
        return -1;

    for (;;) { 
        if (!ReadFile(env->m->service.infile, buf, PROC_BUFSIZE, 
                      &readed, NULL) || readed == 0) 
             break; 
        SwitchToThread();
        if (!WriteFile(env->c->h_stdin[3], buf, readed, 
            &written, NULL))
            break;
        SwitchToThread();
    } 
 
    /* Close the pipe handle so the child process stops reading. */ 
    if (!CloseHandle(env->c->h_stdin[3])) 
      return -1; 
    env->c->h_stdin[3] = INVALID_HANDLE_VALUE;
    return 0;
}


int procrun_redirect(char *program, char **envp, procrun_t *env)
{
    STARTUPINFO si;
    DWORD  id;

    if (!program) {
#ifdef PROCRUN_WINAPP
        MessageBox(NULL, "Service not found", env->m->service.name,
                         MB_OK | MB_ICONERROR);
#else
        fprintf(stderr, "Service not found%s\n", env->m->service.name); 
#endif
        return -1;
    }
    memset(&si, 0, sizeof(STARTUPINFO));
    si.cb = sizeof(si);
    si.dwFlags = STARTF_USESTDHANDLES | STARTF_USESHOWWINDOW;
    si.wShowWindow = SW_HIDE;

    if (procrun_create_pipes(env))
        return -1;
    si.hStdOutput = env->c->h_stdout[1];
    si.hStdError  = env->c->h_stderr[1];
    si.hStdInput  = env->c->h_stdin[0]; 
    
    env->m->envw = make_environment(env->c->env, envp, env->m);
    DBPRINTF1("Creating process %s.\n", program);
    DBPRINTF1("Creating process %s.\n", env->m->argw);
    /* for java.exe merge Arguments and Java options */
    if (env->m->java.jpath) {
        int i, j, len = strlen(env->m->argw) + 8192;
        char *opts[64], *nargw;

        j = make_array(env->m->java.opts, opts, 60, env->m);

        for (i = 0; i < j; i++)
            len += strlen(opts[i]);
       
        nargw = pool_calloc(env->m->pool, len);
        strcpy(nargw, env->m->argw);
        strcat(nargw, " ");
        for (i = 0; i < j; i++) {
            strcat(nargw, opts[i]);
            strcat(nargw, " ");
        }
        strcat(nargw, "-Djava.class.path=");
        if (strchr(program, ' ')) {
            strcat(nargw, "\"");
            strcat(nargw, program);
            strcat(nargw, "\"");
        }
        else
            strcat(nargw, program);
        strcat(nargw, " ");
        strcat(nargw, env->m->java.start_class);
        if (env->m->java.start_param) {
            strcat(nargw, " ");
            strcat(nargw, env->m->java.start_param);
        }
        env->m->argw = nargw;
        program = env->m->java.jpath;
    }
    DBPRINTF2("RUN [%s] %s\n", program, env->m->argw);
    if (env->m->service.account && env->m->service.password &&
        env->m->service.mode == 1) {
        HANDLE user, token;
        if (!LogonUser(env->m->service.account, 
                       NULL, 
                       env->m->service.password, 
                       LOGON32_LOGON_SERVICE,
                       LOGON32_PROVIDER_DEFAULT,
                       &user)) {
            DBPRINTF0(NULL);
            DBPRINTF0("LogonUser failed");
            return -1;
        }

        DuplicateTokenEx(user, 
                         TOKEN_QUERY | TOKEN_DUPLICATE | TOKEN_ASSIGN_PRIMARY, 
                         NULL,
                         SecurityImpersonation,
                         TokenPrimary,
                         &token);
        DBPRINTF0(NULL);

        DBPRINTF2("Launching as %s:%s", env->m->service.account, env->m->service.password);
        ImpersonateLoggedOnUser(token);
        DBPRINTF0(NULL);
        si.lpDesktop = (LPSTR) "Winsta0\\Default";
        if (!CreateProcessAsUser(token,
                                 program,
                                 env->m->argw,
                                 NULL,
                                 NULL,
                                 TRUE,
                                 CREATE_SUSPENDED | CREATE_NEW_CONSOLE | 
                                 CREATE_NEW_PROCESS_GROUP,
                                 env->m->envw,
                                 env->m->service.path,
                                 &si,
                                 &env->c->pinfo)) {

            DBPRINTF1("Error redirecting '%s'\n", program);
            DBPRINTF0(NULL);
            return -1;
        }
    }
    else {
        if (!CreateProcess(program,
                           env->m->argw,
                           NULL,
                           NULL,
                           TRUE,
                           CREATE_SUSPENDED | CREATE_NEW_CONSOLE | 
                           CREATE_NEW_PROCESS_GROUP,
                           env->m->envw,
                           env->m->service.path,
                           &si,
                           &env->c->pinfo)) {
        
            DBPRINTF1("Error redirecting '%s'\n", program);
            DBPRINTF0(NULL);
            return -1;
        }
    }
    pool_handle(env->c->pool, env->c->pinfo.hThread);
    pool_handle(env->c->pool, env->c->pinfo.hProcess);

    SetStdHandle(STD_OUTPUT_HANDLE, env->m->h_stdout[0]);
    SetStdHandle(STD_ERROR_HANDLE, env->m->h_stderr[0]);
    SetStdHandle(STD_INPUT_HANDLE, env->m->h_stdin[0]);  
    CloseHandle(env->c->h_stdout[1]);
    CloseHandle(env->c->h_stderr[1]);
    CloseHandle(env->c->h_stdin[0]); 

    CloseHandle(CreateThread(NULL, 0, stdout_thread, env, 0, &id));     
    CloseHandle(CreateThread(NULL, 0, stderr_thread, env, 0, &id));     
    ResumeThread(env->c->pinfo.hThread);    
    CloseHandle(CreateThread(NULL, 0, wait_thread, env, 0, &id));     

    procrun_write_stdin(env);
    return 0;
}

static int set_service_param(process_t *proc, const char *name,
                             const char *value, int len, int service)
{
    HKEY key;
    char skey[256];
    DWORD err;
    DWORD c, type = REG_SZ;


    if (service == 1) {
        sprintf(skey, PROCRUN_REGKEY_SERVICES, proc->service.name);
        if ((err = RegOpenKeyEx(HKEY_LOCAL_MACHINE, skey,
                            0, KEY_SET_VALUE, &key)) != ERROR_SUCCESS) {
        DBPRINTF2("Failed Creating [%s] name [%s]\n", skey, name);
            return -1;
        }
    }
    else {
        sprintf(skey, PROCRUN_REGKEY_RPARAMS, proc->service.name);
        if ((err = RegCreateKeyEx(HKEY_LOCAL_MACHINE, skey, 0, NULL,
                              0, KEY_SET_VALUE,
                              NULL, &key, &c)) != ERROR_SUCCESS) {
        DBPRINTF0(NULL);
        DBPRINTF2("Failed Creating [%s] [%d]\n", skey, err);
            return -1;
        }
    }
    DBPRINTF2("Creating [%s] name  [%s]\n", skey, name);
    DBPRINTF2("Creating [%s] value [%s]\n", skey, value);
    if (value) {
        if (service == 2)
            type = REG_BINARY;
        else if (len > 0)
            type = REG_MULTI_SZ;
        else
            len = strlen(value);
        err = RegSetValueEx(key, name, 0, type,
            (unsigned char *)value, len);
    }
    else 
        err = RegSetValueEx(key, name, 0, REG_DWORD,
        (unsigned char *)&len, sizeof(int));

    RegCloseKey(key); 
    return (err != ERROR_SUCCESS);
}

/*
 * Process the arguments.
 */
static int process_args(process_t *proc, int argc, char **argv,
                        char **java, char *path)
{
    int arglen = 0;
    char *argp;
    int i,n;

    /* parse command line */
    *java = NULL;
    if (!GetModuleFileName(NULL, path, MAX_PATH - 
                           strlen(proc->service.name) - 7)) {
        return -1;
    } 
    strcat(path, " " PROC_ARG_RUN_SERVICE);
    strcat(path, proc->service.name); 
    for (i = 2; i < argc; i++) {
        DBPRINTF2("Parsing %d [%s]\n", i, argv[i]);
        if (strlen(argv[i]) > 2 && argv[i][0] == '-' && argv[i][1] == '-') {
            argp = &argv[i][2];
            if (STRNI_COMPARE(argp, PROCRUN_PARAMS_IMAGE))
                proc->service.image = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_DESCRIPTION))
                proc->service.description = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_DISPLAY))
                proc->service.display = argv[++i];
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_WORKPATH))
                proc->service.path = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_JVM_OPTS)) {
                proc->java.opts = pool_calloc(proc->pool, strlen(argv[++i]) + 2);
                strcpy(proc->java.opts, argv[i]);
            }
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_JVM))
                *java = argv[++i];
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STDINFILE))
                proc->service.inname = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STDOUTFILE))
                proc->service.outname = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STDERRFILE))
                proc->service.errname = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STARTCLASS))
                proc->java.start_class = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STOPCLASS))
                proc->java.stop_class = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STARTUP)) {
                ++i;
                if (stricmp(argv[i], "auto") == 0)
                    proc->service.startup = SERVICE_AUTO_START;
                else if (stricmp(argv[i], "manual") == 0)
                    proc->service.startup = SERVICE_DEMAND_START;
            }
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_ACCOUNT))
                proc->service.account = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_PASSWORD))
                proc->service.password = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_INSTALL)) {
                strcpy(path, argv[++i]); 
                strcat(path, " " PROC_ARG_RUN_SERVICE);
                strcat(path, proc->service.name);
            }
            else {
                DBPRINTF1("Unrecognized option %s\n", argv[i]);
                break;
            }
        }
        else {
           DBPRINTF1("Not an Command option option %s\n", argv[i]);
           break; 
        }
    }
    if (*java && !strnicmp(*java, "java", 4))
        arglen = strlen(*java) + 1;
    else if (proc->service.name)
        arglen = strlen(proc->service.name) + 1;
    for (n = i; n < argc; n++) {
        arglen += (strlen(argv[n]) + 1);
        if (strchr(argv[n], ' '))
            arglen += 2;
    }
    if (arglen) {
        ++arglen;
        proc->argw = (char *)pool_calloc(proc->pool, arglen);
        if (*java && !strnicmp(*java, "java", 4))
            strcpy(proc->argw, *java);
        else
            strcpy(proc->argw, proc->service.name);
        for (n = i; n < argc; n++) {
            strcat(proc->argw, " ");
            if (strchr(argv[n], ' ')) {
                strcat(proc->argw, "\"");
                strcat(proc->argw, argv[n]);
                strcat(proc->argw, "\"");
            }
            else
                strcat(proc->argw, argv[n]);
            DBPRINTF1("Adding cmdline %s\n", argv[n]);
        }
    }
    if (!proc->service.startup)
        proc->service.startup = SERVICE_AUTO_START;
    return 0;
}

/*
 * Save the parameters in registry
 */
static void save_service_params(process_t *proc, char *java)
{
    int i;

    if (proc->argw)
        set_service_param(proc, PROCRUN_PARAMS_CMDARGS, proc->argw, 0, 0);
    if (proc->service.description) {
        set_service_param(proc, PROCRUN_PARAMS_DESCRIPTION,
                          proc->service.description, 0, 0);
        set_service_param(proc, PROCRUN_PARAMS_DESCRIPTION,
                          proc->service.description, 0, 1);
    }
    if (proc->service.display) {
        set_service_param(proc, PROCRUN_PARAMS_DISPLAY,
                          proc->service.display, 0, 0);
        set_service_param(proc, PROCRUN_PARAMS_DISPLAY,
                          proc->service.display, 0, 1);
    }
    if (proc->service.image)
        set_service_param(proc, PROCRUN_PARAMS_IMAGE,
                          proc->service.image, 0, 0);
    if (proc->service.path)
        set_service_param(proc, PROCRUN_PARAMS_WORKPATH,
                          proc->service.path, 0, 0);
    if (proc->service.inname)
        set_service_param(proc, PROCRUN_PARAMS_STDINFILE,
                          proc->service.inname, 0, 0);
    if (proc->service.outname)
        set_service_param(proc, PROCRUN_PARAMS_STDOUTFILE,
                          proc->service.outname, 0, 0);
    if (proc->service.errname)
        set_service_param(proc, PROCRUN_PARAMS_STDERRFILE,
                          proc->service.errname, 0, 0);
    if (java)
        set_service_param(proc, PROCRUN_PARAMS_JVM,
                          java, 0, 0);
    if (proc->java.start_class)
        set_service_param(proc, PROCRUN_PARAMS_STARTCLASS,
                          proc->java.start_class, 0, 0);
    if (proc->java.stop_class)
        set_service_param(proc, PROCRUN_PARAMS_STOPCLASS,
                          proc->java.stop_class, 0, 0);
    if (proc->service.account)
        set_service_param(proc, PROCRUN_PARAMS_ACCOUNT,
                          proc->service.account, 0, 0);
    if (proc->service.password) {
        unsigned char b[256];
        simple_encrypt(100, proc->service.password, b);
        set_service_param(proc, PROCRUN_PARAMS_PASSWORD,
                          b, 256, 2);
    }
    if (proc->java.opts) {
        int l = strlen(proc->java.opts);
        /* change the string to zero separated for MULYI_SZ */
        for (i = 0; i < l; i ++) {
            if (proc->java.opts[i] == '#')
                proc->java.opts[i] = '\0';
        }
        set_service_param(proc, PROCRUN_PARAMS_JVM_OPTS,
                          proc->java.opts, l + 2, 0);
    }
}

static int procrun_install_service(process_t *proc, int argc, char **argv)
{
    SC_HANDLE service;
    SC_HANDLE manager;
    char path[MAX_PATH+1] = {0};
    char *java = NULL;

    if (!proc->service.name) {
        return -1;
    }

    if (process_args(proc, argc, argv, &java, path)) {
         DBPRINTF0("Installing NT service: process_args failed\n");
         return -1;
    }

    DBPRINTF2("Installing NT service %s %s", path, proc->service.display);

    manager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (!manager) {    
        return -1;
    }

    service = CreateService(manager, 
                            proc->service.name,
                            proc->service.display,
                            SERVICE_ALL_ACCESS,
                            SERVICE_WIN32_OWN_PROCESS,
                            proc->service.startup,
                            SERVICE_ERROR_NORMAL,
                            path,
                            NULL,
                            NULL,
                            SERVICE_DEPENDENCIES,
                            NULL,
                            NULL);
    if (service) {
        CloseServiceHandle(service);
        CloseServiceHandle(manager);
    }
    else {
        DBPRINTF0(NULL);
        CloseServiceHandle(manager);
        return -1;
    }

    /* Save parameters in registry */
    save_service_params(proc,java);

    SetEvent(proc->events[0]);
    return 0;
}

/* Install service on win9x */
static int procrun_install_service9x(process_t *proc, int argc, char **argv)
{
    HKEY        hkey;
    DWORD rv;
    char szPath[MAX_PATH+1] = {0};

    char path[MAX_PATH+1];
    char *display = NULL;
    char *java = NULL;

    DBPRINTF0( "InstallSvc for non-NT\r\n");

    if (!proc->service.name) {
        return -1;
    }

    rv = RegCreateKey(HKEY_LOCAL_MACHINE, "Software\\Microsoft\\Windows"
              "\\CurrentVersion\\RunServices", &hkey);
    if (rv != ERROR_SUCCESS) {
        DBPRINTF0( "Could not open the RunServices registry key\r\n");
        return -1;
    }

    if (!process_args(proc, argc, argv, &java, path)) {
         DBPRINTF0("Installing service: process_args failed\n");
         return -1;
    }

    DBPRINTF1("Installing service %s\n", path);

    rv = RegSetValueEx(hkey, proc->service.name, 0, REG_SZ,
               (unsigned char *) path,
               strlen(path) + 1);
    RegCloseKey(hkey);
    if (rv != ERROR_SUCCESS) {
        DBPRINTF2( "Could not add %s:%s to RunServices Registry Key\r\n",
                  proc->service.name, path);
        return -1;
    }

    strcpy(szPath,
         "SYSTEM\\CurrentControlSet\\Services\\");
    strcat(szPath,proc->service.name);
    rv = RegCreateKey(HKEY_LOCAL_MACHINE, szPath, &hkey);
    if (rv != ERROR_SUCCESS) {
        DBPRINTF1( "Could not create/open the %s registry key\r\n",
            szPath);
        return -1;
    }
    rv = RegSetValueEx(hkey, "ImagePath", 0, REG_SZ,
               (unsigned char *) path,
               strlen(path) + 1);
    if (rv != ERROR_SUCCESS) {
        RegCloseKey(hkey);
        DBPRINTF0( "Could not add ImagePath to our Registry Key\r\n");
        return -1;
    }
    rv = RegSetValueEx(hkey, "DisplayName", 0, REG_SZ,
               (unsigned char *) proc->service.display,
               strlen(proc->service.display) + 1);
    RegCloseKey(hkey);
    if (rv != ERROR_SUCCESS) {
        DBPRINTF0( "Could not add DisplayName to our Registry Key\r\n");
        return -1;
    }

    /* Save parameters in registry */
    save_service_params(proc,java);

    return 0;
}

static int procrun_update_service(process_t *proc, int argc, char **argv)
{
    SC_HANDLE service;
    SC_HANDLE manager;
    char *argp;
    char path[MAX_PATH+1];
    char *java = NULL;
    int arglen = 0;

    int i, n;
    if (!proc->service.name) {
        return -1;
    }
    if (!GetModuleFileName(NULL, path, MAX_PATH - 
                           strlen(proc->service.name) - 7)) {
        DBPRINTF0(NULL);
        return -1;
    }
    strcat(path, " " PROC_ARG_RUN_SERVICE);
    strcat(path, proc->service.name);
    DBPRINTF1("Updating service %s\n", path);

    manager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (!manager) {    
        DBPRINTF0(NULL);
        return -1;
    }
    service = OpenService(manager, proc->service.name, SERVICE_ALL_ACCESS);
    if (!service) {
        DBPRINTF0(NULL);
        CloseServiceHandle(manager);
        return -1;
    }
    proc->service.startup = SERVICE_NO_CHANGE;
    proc->service.image = NULL;
    proc->service.description = NULL;
    proc->service.path = NULL;
    proc->service.inname = NULL;
    proc->service.outname = NULL;
    proc->service.errname = NULL;
    proc->java.start_class = NULL;
    proc->java.stop_class = NULL;
    proc->java.opts = NULL;
    proc->service.account  = NULL;
    proc->service.password = NULL;
    proc->service.display = NULL;
    proc->argw = NULL;
    /* parse command line */
    for (i = 2; i < argc; i++) {
        if (strlen(argv[i]) > 2 && argv[i][0] == '-' && argv[i][1] == '-') {
            argp = &argv[i][2];
            if (STRNI_COMPARE(argp, PROCRUN_PARAMS_IMAGE))
                proc->service.image = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_DESCRIPTION))
                proc->service.description = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_DISPLAY))
                proc->service.display = argv[++i];
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_WORKPATH))
                proc->service.path = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_JVM_OPTS)) {
                proc->java.opts = pool_calloc(proc->pool, strlen(argv[++i]) + 2);
                strcpy(proc->java.opts, argv[i]);
            }
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_JVM))
                java = argv[++i];
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STDINFILE))
                proc->service.inname = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STDOUTFILE))
                proc->service.outname = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STDERRFILE))
                proc->service.errname = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STARTCLASS))
                proc->java.start_class = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STOPCLASS))
                proc->java.stop_class = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_STARTUP)) {
                ++i;
                if (stricmp(argv[i], "auto") == 0)
                    proc->service.startup = SERVICE_AUTO_START;
                else if (stricmp(argv[i], "manual") == 0)
                    proc->service.startup = SERVICE_DEMAND_START;
            }
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_ACCOUNT))
                proc->service.account = pool_strdup(proc->pool, argv[++i]);
            else if (STRNI_COMPARE(argp, PROCRUN_PARAMS_PASSWORD))
                proc->service.password = pool_strdup(proc->pool, argv[++i]);
            else
                break;
        }
        else
            break;
    }

    if (i < argc) {
        if (java && !strnicmp(java, "java", 4))
            arglen = strlen(java) + 1;
        else if (proc->service.name)
            arglen = strlen(proc->service.name) + 1;
        for (n = i; n < argc; n++) {
            arglen += strlen(argv[n]) + 1;
            if (strchr(argv[n], ' '))
                arglen += 2;
        }
        if (arglen) {
            ++arglen;
            proc->argw = (char *)pool_calloc(proc->pool, arglen);
            if (java && !strnicmp(java, "java", 4))
                strcpy(proc->argw, java);
            else
                strcpy(proc->argw, proc->service.name);
            for (n = i; n < argc; n++) {
                strcat(proc->argw, " ");
                if (strchr(argv[n], ' ')) {
                    strcat(proc->argw, "\"");
                    strcat(proc->argw, argv[n]);
                    strcat(proc->argw, "\"");
                }
                else
                    strcat(proc->argw, argv[n]);
                DBPRINTF1("Adding cmdline %s\n", argv[n]);
            }
        }
    }
    ChangeServiceConfig(service,
                        SERVICE_NO_CHANGE,
                        proc->service.startup,
                        SERVICE_NO_CHANGE,
                        NULL,
                        NULL,
                        NULL,
                        NULL,
                        NULL,
                        NULL,
                        proc->service.display);

    CloseServiceHandle(service);
    CloseServiceHandle(manager);
    
    save_service_params(proc,java);

    SetEvent(proc->events[0]);
    return 0;
}

static int procrun_delete_service(process_t *proc)
{
    SC_HANDLE service;
    SC_HANDLE manager;
    SERVICE_STATUS status;
    char skey[MAX_PATH+1];
    DWORD err;

    if (!proc->service.name) {
        return -1;
    }
    DBPRINTF1("Deleting service %s", proc->service.name);

    sprintf(skey, PROCRUN_REGKEY_RSERVICES, proc->service.name);
    if ((err = SHDeleteKey(HKEY_LOCAL_MACHINE, skey)) != ERROR_SUCCESS) {
        DBPRINTF2("Failed Deleting [%s] [%d]\n", skey, err);
    }

    manager = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);
    if (!manager) {    
        DBPRINTF0(NULL);
        return -1;
    }
    service = OpenService(manager, proc->service.name, SERVICE_ALL_ACCESS);
    
    if (service) {
        BOOL ss = FALSE;
        QueryServiceStatus(service, &status);
        if (status.dwCurrentState != SERVICE_RUNNING)
            ss = DeleteService(service);
        else {
            /* Stop the service */
            if (ControlService(service, SERVICE_CONTROL_STOP, &status)) {
                Sleep(1000);
                while (QueryServiceStatus(service, &status)) {
                    if (status.dwCurrentState == SERVICE_STOP_PENDING)
                        Sleep(1000);
                    else
                        break;
                }
            }
            QueryServiceStatus(service, &status);
            if (status.dwCurrentState != SERVICE_RUNNING)
                ss = DeleteService(service);
        }
        CloseServiceHandle(service);
        CloseServiceHandle(manager);
        if (!ss)
            return -1;
    }
    else {
        DBPRINTF0(NULL);
        CloseServiceHandle(manager);
        return -1;
    }
    SetEvent(proc->events[0]);
    return 0;
}
 
/* remove service (non NT) stopping it looks ugly!!! so we let it run. */
static int procrun_delete_service9x(process_t *proc)
{
    HKEY hkey;
    DWORD rv;
 
    if (!proc->service.name) {
        return -1;
    }
    DBPRINTF1("Deleting service %s", proc->service.name);

    rv = RegOpenKey(HKEY_LOCAL_MACHINE,
        "Software\\Microsoft\\Windows\\CurrentVersion\\RunServices",
        &hkey);
    if (rv != ERROR_SUCCESS) {
        DBPRINTF0( "Could not open the RunServices registry key.\r\n");
        return -1;
    }
    rv = RegDeleteValue(hkey, proc->service.name);
    RegCloseKey(hkey);
    if (rv != ERROR_SUCCESS)
        DBPRINTF0( "Could not delete the RunServices entry.\r\n");
 
    rv = RegOpenKey(HKEY_LOCAL_MACHINE,
        "SYSTEM\\CurrentControlSet\\Services", &hkey);
    if (rv != ERROR_SUCCESS) {
        DBPRINTF0( "Could not open the Services registry key.\r\n");
        return -1;
    }
    rv = RegDeleteKey(hkey, proc->service.name);
    RegCloseKey(hkey);
    if (rv != ERROR_SUCCESS) {
        DBPRINTF0( "Could not delete the Services registry key.\r\n");
        return -1;
    }
    return 0;
}

int report_service_status(DWORD dwCurrentState,
                          DWORD dwWin32ExitCode,
                          DWORD dwWaitHint,
                          process_t *proc)
{
   static DWORD dwCheckPoint = 1;
   BOOL fResult = TRUE;

   if (proc->service.mode && proc->service.h_status) {
      if (dwCurrentState == SERVICE_START_PENDING)
         proc->service.status.dwControlsAccepted = 0;
      else
         proc->service.status.dwControlsAccepted = SERVICE_ACCEPT_STOP;

      proc->service.status.dwCurrentState = dwCurrentState;
      proc->service.status.dwWin32ExitCode = dwWin32ExitCode;
      proc->service.status.dwWaitHint = dwWaitHint;

      if ((dwCurrentState == SERVICE_RUNNING) ||
          (dwCurrentState == SERVICE_STOPPED))
         proc->service.status.dwCheckPoint = 0;
      else
        proc->service.status.dwCheckPoint = dwCheckPoint++;
      if (!(fResult = SetServiceStatus(
                        proc->service.h_status, &proc->service.status))) {
         DBPRINTF0("SetServiceStatus\tFailed\n");
         DBPRINTF0(NULL);
      }
   }
   return fResult;
}

void WINAPI service_ctrl(DWORD dwCtrlCode)
{
    switch (dwCtrlCode) {
        case SERVICE_CONTROL_STOP:
            report_service_status(SERVICE_STOP_PENDING, NO_ERROR, 0,
                                  g_env->m);
            DBPRINTF0("ServiceCtrl STOP\n");
            SetEvent(g_env->m->events[0]);
            return;
        case SERVICE_CONTROL_INTERROGATE:
        break;
        default:
        break;
   }

   report_service_status(g_env->m->service.status.dwCurrentState,
                         NO_ERROR, 0, g_env->m);
}

BOOL WINAPI console_ctrl(DWORD dwCtrlType)
{
    switch (dwCtrlType) {
        case CTRL_BREAK_EVENT:
            DBPRINTF0("Captured CTRL+BREAK Event\n");
            SetEvent(g_env->m->events[0]);
            return TRUE;
        case CTRL_C_EVENT:
            DBPRINTF0("Captured CTRL+C Event\n");
            SetEvent(g_env->m->events[0]);
            return TRUE;
        case CTRL_CLOSE_EVENT:
            DBPRINTF0("Captured CLOSE Event\n");
            SetEvent(g_env->m->events[0]);
            return TRUE;
        case CTRL_SHUTDOWN_EVENT:
            DBPRINTF0("Captured SHUTDOWN Event\n");
            SetEvent(g_env->m->events[0]);
            return TRUE;
        break;

   }
   return FALSE;
}

DWORD WINAPI java_thread(LPVOID param)
{
    DWORD rv;
    procrun_t *env = (procrun_t *)param;
    
    if (!param)
        return -1;
    rv = procrun_init_jvm(env->m);
    jni_detach(env->m);

    SetEvent(env->m->events[0]);
    return rv;
}

int service_main(int argc, char **argv)
{
    DWORD fired = 0;
    int   rv = -1;
    
    
    g_env->m->service.status.dwServiceType      = SERVICE_WIN32_OWN_PROCESS;
    g_env->m->service.status.dwCurrentState     = SERVICE_START_PENDING; 
    g_env->m->service.status.dwControlsAccepted = SERVICE_ACCEPT_STOP | SERVICE_ACCEPT_PAUSE_CONTINUE; 
    g_env->m->service.status.dwWin32ExitCode    = 0; 
    g_env->m->service.status.dwServiceSpecificExitCode = 0; 
    g_env->m->service.status.dwCheckPoint       = 0; 
    g_env->m->service.status.dwWaitHint         = 0; 

    g_env->m->service.status.dwServiceSpecificExitCode = 0;
    if (g_env->m->service.mode) {
        g_env->m->service.h_status = RegisterServiceCtrlHandler(g_env->m->service.name,
                                                                     service_ctrl);
        if (!g_env->m->service.h_status)
            goto cleanup;
    }
    report_service_status(SERVICE_START_PENDING, NO_ERROR, 3000,
                          g_env->m);
    if (g_env->m->java.dll) {
        if (g_env->m->java.jpath == NULL) {
            DWORD id;
            HANDLE jh;
            jh = CreateThread(NULL, 0, java_thread, g_env, 0, &id);
            pool_handle(g_env->m->pool, jh);
            rv = 0;
        }
        else {
            if ((rv = procrun_init_jvm(g_env->m)) == 0) {
                rv = procrun_redirect(g_env->m->service.image,
                                      g_env->m->envp, g_env);    
            }
        }
    } 
    else {
        rv = procrun_redirect(g_env->m->service.image,
                              g_env->m->envp, g_env);    
    }
    if (rv == 0) {
        report_service_status(SERVICE_RUNNING, NO_ERROR, 0,
                              g_env->m);
        DBPRINTF2("Service %s Started %d\n", g_env->m->service.name, rv);

    }
    /* Wait for shutdown event or child exit */
    while (rv == 0) {
        fired = WaitForMultipleObjects(4, g_env->m->events,
                                       FALSE, INFINITE); 
        
        if (fired == WAIT_OBJECT_0 ||
            fired == (WAIT_OBJECT_0 + 1) ||
            fired == WAIT_TIMEOUT)
            break;
    }

cleanup:
    DBPRINTF1("Stoping Service %s\n", g_env->m->service.name);
    report_service_status(SERVICE_STOP_PENDING, NO_ERROR, 3000,
                          g_env->m);
    procrun_destroy_jvm(g_env->m);

    inject_exitprocess(&g_env->c->pinfo);
    report_service_status(SERVICE_STOPPED, 0, 0,
                          g_env->m);

    DBPRINTF0("Setting Shutdown Event 0\n");
    SetEvent(g_env->m->events[0]);
    return rv;
}


int procrun_main(int argc, char **argv, char **envp, procrun_t *env)
{
    int i, mode = 0;
    char event[64];
    DWORD fired;
    int   rv = -1;
    OSVERSIONINFO osver;

    SERVICE_TABLE_ENTRY dispatch_table[] = {
        {NULL, NULL},
        {NULL, NULL}
    };

   osver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
    if (GetVersionEx(&osver)) {
        if (osver.dwPlatformId >= VER_PLATFORM_WIN32_NT)
            g_is_windows_nt = 1;
    }
    DBPRINTF1("OS Version %d", g_is_windows_nt);

    SetConsoleCtrlHandler((PHANDLER_ROUTINE)console_ctrl, TRUE);
    env->m->pool = pool_create();
    env->c->pool = pool_create();
    for (i = 0; i < 4; i ++) {
        env->m->h_stdin[i]   = INVALID_HANDLE_VALUE;
        env->m->h_stdout[i]  = INVALID_HANDLE_VALUE;
        env->m->h_stderr[i]  = INVALID_HANDLE_VALUE;
        env->c->h_stdin[i]  = INVALID_HANDLE_VALUE;
        env->c->h_stdout[i] = INVALID_HANDLE_VALUE;
        env->c->h_stderr[i] = INVALID_HANDLE_VALUE;
    }

    env->m->pinfo.dwProcessId = GetCurrentProcessId();
    env->m->pinfo.dwThreadId = GetCurrentThreadId();
    env->m->pinfo.hProcess = GetCurrentProcess();
    env->m->pinfo.hThread = GetCurrentThread();
    env->m->envp = envp;

    SetProcessShutdownParameters(0x300, SHUTDOWN_NORETRY);
    if (argc > 1)
        mode = parse_args(argc, argv, env->m);
    else /* nothing to do. exit... */
        return 0;
    
    procrun_readenv(env->m, envp);
    
    /* Add the pid to the client env */
    procrun_addenv(PROCRUN_ENV_PPID, NULL, (int)GetCurrentProcessId(),
                   env->m);
    
    procrun_service_params(env->m);
    /* Check if we have a JVM */
    if (procrun_load_jvm(env->m, mode) < 0)
        goto cleanup;
    sprintf(event, "PROC_SHUTDOWN_EVENT%d", GetCurrentProcessId());
    env->m->events[0] = CreateEvent(NULL, TRUE, FALSE, event);
    sprintf(event, "PROC_EXITWAIT_EVENT%d", GetCurrentProcessId());
    env->m->events[1] = CreateEvent(NULL, TRUE, FALSE, event);
    sprintf(event, "PROC_CLOSESTDOUT_EVENT%d", GetCurrentProcessId());
    env->m->events[2] = CreateEvent(NULL, TRUE, FALSE, event);
    sprintf(event, "PROC_CLOSESTDERR_EVENT%d", GetCurrentProcessId());
    env->m->events[3] = CreateEvent(NULL, TRUE, FALSE, event);
    pool_handle(env->m->pool, env->m->events[0]);
    pool_handle(env->m->pool, env->m->events[1]);
    pool_handle(env->m->pool, env->m->events[2]);
    pool_handle(env->m->pool, env->m->events[3]);

    DBPRINTF2("running in mode %d %p\n", mode, env->m->java.dll);
    DBPRINTF1("Events %p\n", env->m->events[0]);

    switch (mode) {
        case PROCRUN_CMD_TEST_SERVICE:
#ifdef PROCRUN_WINAPP
        case PROCRUN_CMD_GUIT_SERVICE:
        case PROCRUN_CMD_GUID_SERVICE:
            {
                DWORD gi;
                LoadLibrary("riched32.dll");
                CreateThread(NULL, 0, gui_thread, g_env, 0, &gi);
            }
#else
            if (env->m->service.description)
                SetConsoleTitle(env->m->service.description);
            else if (env->m->service.name)
                SetConsoleTitle(env->m->service.name);
#endif
#ifdef _DEBUG
            debug_process(argc, argv, env->m);
#endif
            service_main(argc, argv);
            break;
        case PROCRUN_CMD_INSTALL_SERVICE:
            if (g_is_windows_nt)
                rv = procrun_install_service(env->m, argc, argv);
            else
                rv = procrun_install_service9x(env->m, argc, argv);
            break;
        case PROCRUN_CMD_UPDATE_SERVICE:
            if (g_is_windows_nt)
                rv = procrun_update_service(env->m, argc, argv);
            else {
              // rv = procrun_update_service9x(env->m, argc, argv);
                rv = -1;
                DBPRINTF0("UPDATE SERVICE is unimplemented on 9x for now");
            }
            break;
        case PROCRUN_CMD_DELETE_SERVICE:
            if (g_is_windows_nt)
                rv = procrun_delete_service(env->m);
            else
                rv = procrun_delete_service9x(env->m);
            break;
        case PROCRUN_CMD_STOP_SERVICE:
            rv = -1;
            DBPRINTF0("STOP SERVICE is unimplemented for now");
            break;
        case PROCRUN_CMD_RUN_SERVICE:
            debug_process(argc, argv, env->m);
            env->m->service.mode = 1;
            dispatch_table[0].lpServiceName = env->m->service.name;
            dispatch_table[0].lpServiceProc = (LPSERVICE_MAIN_FUNCTION)service_main;
            rv = (StartServiceCtrlDispatcher(dispatch_table) == FALSE);
            break;
#ifdef PROCRUN_WINAPP
        case PROCRUN_CMD_GUID_PROCESS:
            {
                DWORD gi;
                ac_use_try = 1;
                LoadLibrary("riched32.dll");
                CreateThread(NULL, 0, gui_thread, g_env, 0, &gi);
            }
            env->m->service.name = argv[0];
            env->m->service.description = argv[2];
            rv = procrun_redirect(argv[2], envp, env);
            break;

#endif
        default:
            /* Run the process 
             * We can be master or called by env.
             */
            rv = procrun_redirect(argv[1], envp, env);
            break;
    }
    
    /* Wait for shutdown event or child exit */
    while (rv == 0) {
        fired = WaitForMultipleObjects(4, env->m->events,
                                       FALSE, INFINITE); 
        
        if (fired == WAIT_OBJECT_0 ||
            fired == (WAIT_OBJECT_0 + 1) ||
            fired == WAIT_TIMEOUT)
            break;
    }
    DBPRINTF1("Exiting from mode %d\n", mode);
    /* clean up... */
cleanup:

#ifdef PROCRUN_WINAPP
    if (ac_main_hwnd)
        ac_show_try_icon(ac_main_hwnd, NIM_DELETE, NULL);
#endif
    procrun_destroy_jvm(env->m);
    inject_exitprocess(&env->c->pinfo);
    i = pool_destroy(env->m->pool);
    i = pool_destroy(env->c->pool);
    if (g_proc_stderr_file)
        close(g_proc_stderr_file);
    return 0;
}

static procrun_t *alloc_environment()
{

    procrun_t *env = (procrun_t *)malloc(sizeof(procrun_t));
    if (env) {
        env->m = (process_t *)calloc(1, sizeof(process_t));
        env->c = (process_t *)calloc(1, sizeof(process_t));
    }
    return env;
}

static void free_environment(procrun_t *env)
{
    if (env) {
        free(env->m);
        free(env->c);
        free(env);
        env = NULL;
    }
}

#if defined(PROCRUN_WINAPP)
#pragma message("Compiling WIN32 Application mode")

int WINAPI WinMain(HINSTANCE hInstance,
                    HINSTANCE hPrevInstance,
                    LPSTR lpCmdLine,
                    int nCmdShow)
{
    int rv;
    procrun_t *env = alloc_environment();
    g_proc_mode = PROCRUN_MODE_WINAPP;
    g_env = env;
    ac_instance = hInstance;

    rv = procrun_main(__argc, __argv, _environ, env);

    free_environment(env);
    return rv;
}
#elif defined(PROCRUN_CONSOLE)
#pragma message("Compiling CONSOLE Application mode")

void __cdecl main(int argc, char **argv)
{

    procrun_t *env = alloc_environment();
    g_proc_mode = PROCRUN_MODE_CONSOLE;
    g_env = env;

    procrun_main(argc, argv, _environ, env);

    free_environment(env);
}
#elif defined(PROCRUN_WINDLL)
#pragma message("Compiling DLL Application mode")


BOOL WINAPI DllMain(HINSTANCE hInst,
                    ULONG ulReason,
                    LPVOID lpReserved)
{ 

    switch (ulReason) {
        case DLL_PROCESS_ATTACH:
            g_env = NULL;
        break;
        case DLL_PROCESS_DETACH:
            free_environment(g_env);
        break;
        default:
        break;
    } 
    return TRUE;     
}

__declspec(dllexport) void InstallService(const char *service_name,
                                          const char *install,
                                          const char *image_path,
                                          const char *display_name,
                                          const char *description)
{
    int argc = 0;
    char *argv[12];
    char b[MAX_PATH];

    procrun_t *env = alloc_environment();
    g_proc_mode = PROCRUN_MODE_WINDLL;
    g_env = env;
    
    argv[argc++] = "PROCRUN.DLL";
    strcpy(b, PROC_ARG_INSTALL_SERVICE);
    strcat(b, service_name);
    argv[argc++] = b;
    argv[argc++] = "--" PROCRUN_PARAMS_IMAGE;
    argv[argc++] = (char *)image_path;
    argv[argc++] = "--" PROCRUN_PARAMS_INSTALL;
    argv[argc++] = (char *)install;
    argv[argc++] = "--" PROCRUN_PARAMS_DISPLAY;
    argv[argc++] = (char *)display_name;
    argv[argc++] = "--" PROCRUN_PARAMS_DESCRIPTION;
    argv[argc++] = (char *)description;
    
    procrun_main(argc, argv, _environ, env);

    free_environment(env);
    g_env = NULL;
}

__declspec(dllexport) void UpdateService(const char *service_name,
                                         const char *param,
                                         const char *value)
{
    int argc = 0;
    char *argv[4];
    char b[MAX_PATH], p[MAX_PATH];

    procrun_t *env = alloc_environment();
    g_proc_mode = PROCRUN_MODE_WINDLL;
    g_env = env;
    
    argv[argc++] = "PROCRUN.DLL";
    strcpy(b, PROC_ARG_UPDATE_SERVICE);
    strcat(b, service_name);
    strcpy(p, "--");
    strcat(p, param);
    argv[argc++] = b;
    argv[argc++] = p;
    argv[argc++] = (char *)value;
    
    procrun_main(argc, argv, _environ, env);

    free_environment(env);
    g_env = NULL;
}

__declspec(dllexport) void RemoveService(const char *service_name)
{
    int argc = 0;
    char *argv[4];
    char b[MAX_PATH];

    procrun_t *env = alloc_environment();
    g_proc_mode = PROCRUN_MODE_WINDLL;
    g_env = env;
    
    argv[argc++] = "PROCRUN.DLL";
    strcpy(b, PROC_ARG_DELETE_SERVICE);
    strcat(b, service_name);
    argv[argc++] = b;
    procrun_main(argc, argv, _environ, env);

    free_environment(env);
    g_env = NULL;
}
#else
#error Unknown application mode
#endif

